#!/usr/bin/env python
import unittest
import xypath.contrib.excel

class Test_Contrib(unittest.TestCase):
    def test_excel_nums(self):
        num = xypath.contrib.excel.excel_column_number
        let = xypath.contrib.excel.excel_column_label
        cases = [
        ['A', 1],
        ['Z', 26],
        ['AA', 27],
        ['AZ', 52],
        ['BA', 53],
        ['AAA', 703],
        ]
        for case in cases:
            assert num(case[0]) == case[1], num(case[0])
            assert let(case[1]) == case[0]
            assert num(case[0], index=0) == case[1] - 1



#!/usr/bin/env python
import sys
sys.path.append('xypath')
import xypath
import tcore


class TestFill(tcore.TCore):

    def test_fill_down_without_termination(self):
        world_cell = self.table.filter('WORLD').assert_one()
        filled_down = world_cell.fill(xypath.DOWN)
        assert world_cell not in filled_down
        self.assertEqual(264, len(filled_down))

    def test_fill_right_without_termination(self):
        world_cell = self.table.filter('WORLD').assert_one()
        filled_right = world_cell.fill(xypath.RIGHT)
        assert world_cell not in filled_right
        self.assertEqual(14, len(filled_right))

    def test_fill_down_with_termination(self):
        world_cell = self.table.filter('WORLD').assert_one()
        stop_cell = self.table.filter('Sub-Saharan Africa').assert_one()

        filled_down = world_cell.fill(
            xypath.DOWN,
            stop_before=lambda cell: cell == stop_cell)
        assert world_cell not in filled_down
        assert stop_cell not in filled_down  # function should be exclusive
        self.assertEqual(5, len(filled_down))

    def test_fill_right_with_termination(self):
        index_cell = self.table.filter('Index').assert_one()
        stop_cell = self.table.filter('Notes').assert_one()

        filled_down = index_cell.fill(
            xypath.RIGHT,
            stop_before=lambda cell: cell == stop_cell)
        assert index_cell not in filled_down
        assert stop_cell not in filled_down  # function should be exclusive
        self.assertEqual(2, len(filled_down))



import tcore
import xypath.xypath as xypath # no privacy

import unittest

from copy import copy

class Test_Lookup(unittest.TestCase):
    @classmethod
    def setup_class(cls):
        cls.wpp_filename = tcore.get_fixture_filename("lookup.csv")
        cls.messy, cls.table = tcore.get_messytables_fixture(cls.wpp_filename)

    def test_cell_lookup(self):
        print self.table
        v = self.table.filter("V")._cell
        a = self.table.filter("A")
        b = self.table.filter("B")
        try:
            v.lookup(a, xypath.UP).shift(1,0).value == '5'
        except xypath.LookupConfusionError:
            pass
        else:
            raise AssertionError
        assert v.lookup(a, xypath.UP, strict=True).shift(1,0).value == '4'
        assert v.lookup(a, xypath.DOWN).shift(1,0).value == '7'
        assert v.lookup(a, xypath.DOWN, strict=True).shift(1,0).value == '9'
        assert v.lookup(b, xypath.LEFT).shift(1,0).value == '0'
        try:
            v.lookup(b, xypath.LEFT, strict=True) # should fail
        except xypath.NoLookupError:
            pass
        else:
            raise AssertionError
        assert v.lookup(b, xypath.RIGHT).shift(1,0).value == '1'
        assert v.lookup(b, xypath.RIGHT, strict=True).shift(1,0).value == '4'

class XYCellTests(unittest.TestCase):

    def test_cell_equality(self):
        """
        test_cell_equality: required for expected set behaviour
        """
        table = xypath.Table()

        cell_a = xypath._XYCell("foo", 1, 3, table)
        also_cell_a = xypath._XYCell("foo", 1, 3, table)

        self.assertIsNot(cell_a, also_cell_a)
        self.assertEqual(cell_a, also_cell_a)

    def test_cell_identity(self):
        """
        test_cell_identity: given two cells which were instantiated separately
        but *actually* belong to the same logical cell (table, co-ordinate,
        value), ensure that their identity (as far as the 'set' type is
        concerned) is maintained.
        """

        table = xypath.Table()

        cell_a = xypath._XYCell("foo", 1, 3, table)
        also_cell_a = xypath._XYCell("foo", 1, 3, table)

        self.assertIsNot(cell_a, also_cell_a)
        self.assertEqual(hash(cell_a), hash(also_cell_a))

        self.assertEqual(1, len(set([cell_a, also_cell_a])))

    def test_cell_copy(self):
        table = xypath.Table()

        cell_a = xypath._XYCell("foo", 1, 3, table)
        not_cell_a = copy(cell_a)
        # Before mutating "not_cell_a", ensure that it is first equal
        self.assertEqual(hash(cell_a), hash(not_cell_a))


    def test_cell_identity_not_equal_different_x(self):

        table = xypath.Table()
        cell_a = xypath._XYCell("foo", 1, 3, table)
        not_cell_a = copy(cell_a)
        not_cell_a.x = 2
        self.assertNotEqual(hash(cell_a), hash(not_cell_a))
        self.assertEqual(2, len(set([cell_a, not_cell_a])))

    def test_cell_identity_not_equal_different_y(self):
        table = xypath.Table()
        cell_a = xypath._XYCell("foo", 1, 3, table)
        not_cell_a = copy(cell_a)
        not_cell_a.y = 2
        self.assertNotEqual(hash(cell_a), hash(not_cell_a))
        self.assertEqual(2, len(set([cell_a, not_cell_a])))


    def test_cell_identity_not_equal_different_tables(self):
        table = xypath.Table()
        cell_a = xypath._XYCell("foo", 1, 3, table)

        other_table = xypath.Table()
        self.assertNotEqual(hash(table), hash(other_table))

        not_cell_a = copy(cell_a)
        not_cell_a.table = other_table

        self.assertNotEqual(hash(cell_a), hash(not_cell_a))
        self.assertEqual(2, len(set([cell_a, not_cell_a])))

    def test_cell_shift_takes_tuples(self):
        table = xypath.Table()
        cell = xypath._XYCell("foo", 1, 3, table)
        table.add(cell)
        cell.shift([0,0])




import tcore
import xypath
from textwrap import dedent
from cStringIO import StringIO


class TestPPrint(tcore.TCore):

    def test_pprint(self):
        """
        test_pprint: basic check of pprint output against hard-coded string
        """

        EXPECTED_TABLE = dedent('''
              /  C
            ---  -------------------------
            274  American Samoa
            275  Cook Islands
            276  French Polynesia
            277  Niue
            278  Samoa
            279  Tokelau
            280  Tonga
            281  Tuvalu
            282  Wallis and Futuna Islands
            ''').lstrip('\n')

        cells = self.table.filter('Polynesia').fill(xypath.DOWN)
        stream = StringIO()
        cells.pprint(stream=stream)
        self.assertEqual(EXPECTED_TABLE, stream.getvalue())
        
    def test_pprint_removed_cell(self):
        """
        test_pprint_removed_cell: check that pprint's output is rectangular
        """

        EXPECTED_TABLE = dedent('''
              /  C
            ---  -------------------------
            274  American Samoa
            275  Cook Islands
            276  French Polynesia
            277  Niue
            278  Samoa
            279  Tokelau
            280  /
            281  Tuvalu
            282  Wallis and Futuna Islands
            ''').lstrip('\n')

        cells = self.table.filter('Polynesia').fill(xypath.DOWN)
        cells_without_tonga = cells - cells.filter('Tonga')

        stream = StringIO()
        cells_without_tonga.pprint(stream=stream)
        self.assertEqual(EXPECTED_TABLE, stream.getvalue())
                
    def test_pprint_removed_cell_collapsed(self):
        """
        test_pprint_removed_cell_collapsed: check that a collapsed row works
        """

        EXPECTED_TABLE = dedent('''
              /  C
            ---  -------------------------
            274  American Samoa
            275  Cook Islands
            276  French Polynesia
            277  Niue
            278  Samoa
            279  Tokelau
            281  Tuvalu
            282  Wallis and Futuna Islands
            ''').lstrip('\n')

        cells = self.table.filter('Polynesia').fill(xypath.DOWN)
        cells_without_tonga = cells - cells.filter('Tonga')
        stream = StringIO()
        cells_without_tonga.pprint(collapse_empty=True, stream=stream)
        self.assertEqual(EXPECTED_TABLE, stream.getvalue())
        
    def test_extrude(self):
        """
        test_extrude: check 2-to-the-left extrusion against hard-coded string
        """

        EXPECTED_TABLE = dedent('''
            /  C                          D      E
          ---  -------------------------  ---  ---
          274  American Samoa                   16
          275  Cook Islands                    184
          276  French Polynesia                258
          277  Niue                            570
          278  Samoa                           882
          279  Tokelau                         772
          280  Tonga                           776
          281  Tuvalu                          798
          282  Wallis and Futuna Islands       876
          ''').lstrip('\n')

        # Below the word "Polynesia", there are a set of islands
        cells = self.table.filter('Polynesia').fill(xypath.DOWN)
        # Right of this column, there is an empty column followed by numbers.
        cells = cells.extrude(2, 0)

        stream = StringIO()
        cells.pprint(stream=stream)

        self.assertEqual(EXPECTED_TABLE, stream.getvalue())

    def test_pprint_collapsed_column(self):
        """
        test_pprint_collapsed_column: check empty column removal
        """

        # Column D is removed because it contains no itmes with bool(x) == True

        EXPECTED_TABLE = dedent('''
              /  C                            E
            ---  -------------------------  ---
            274  American Samoa              16
            275  Cook Islands               184
            276  French Polynesia           258
            277  Niue                       570
            278  Samoa                      882
            279  Tokelau                    772
            280  Tonga                      776
            281  Tuvalu                     798
            282  Wallis and Futuna Islands  876
            ''').lstrip('\n')

        # Below the word "Polynesia", there are a set of islands
        cells = self.table.filter('Polynesia').assert_one().fill(xypath.DOWN)
        # Right of this column, there is an empty column followed by numbers.
        cells = cells.extrude(2, 0)

        stream = StringIO()
        # Collapse empty removes the empty column
        cells.pprint(collapse_empty=True, stream=stream)
        self.assertEqual(EXPECTED_TABLE, stream.getvalue())


    def test_pprint_collapsed_column_and_row(self):
        """
        test_pprint_collapsed_column_and_row: check row/column removal
        """

        # Column D is removed because it contains no itmes with bool(x) == True

        EXPECTED_TABLE = dedent('''
              /  C                            E
            ---  -------------------------  ---
            274  American Samoa              16
            275  Cook Islands               184
            276  French Polynesia           258
            277  Niue                       570
            278  Samoa                      882
            279  Tokelau                    772
            281  Tuvalu                     798
            282  Wallis and Futuna Islands  876
            ''').lstrip('\n')

        # Below the word "Polynesia", there are a set of islands
        cells = self.table.filter('Polynesia').assert_one().fill(xypath.DOWN)
        # Right of this column, there is an empty column followed by numbers.
        cells = cells.extrude(2, 0)

        tonga = cells.filter("Tonga")
        # Remove cells in the same row as tonga
        cells -= cells.same_row(tonga)

        stream = StringIO()
        # Collapse empty removes the empty column
        cells.pprint(collapse_empty=True, stream=stream)
        self.assertEqual(EXPECTED_TABLE, stream.getvalue())



#!/usr/bin/env python
import sys

import unittest
sys.path.append('xypath')
import xypath
import messytables
try:
    import hamcrest
except ImportError:
    hamcrest = None
import re
import tcore


class TestFilter(tcore.TCore):
    def test_bag_function(self):
        self.table.filter("Country code").filter(lambda b: b.shift(x=0, y=0).value=="Country code").assert_one()
    def test_basic_vert(self):
        r = repr(self.table.filter(lambda b: b.x == 2))
        self.assertIn("WORLD", r)
        self.assertNotIn("Country code", r)

    def test_basic_horz(self):
        r = repr(self.table.filter(lambda b: b.y == 16))
        self.assertNotIn("WORLD", r)
        self.assertIn("Country code", r)

    def test_text_match_string(self):
        self.table.filter("Country code").assert_one()

    def test_text_exact_match_lambda(self):
        self.table.filter(lambda b: b.value == 'Country code').assert_one()

    def test_text_exact_match_hamcrest(self):
        if hamcrest is None:
            raise unittest.SkipTest("Requires Hamcrest")
        self.table.filter(hamcrest.equal_to("Country code")).assert_one()

    def test_regex_match(self):
        self.assertEqual(
            2, len(self.table.filter(re.compile(r'.... developed regions$'))))

    def test_regex_not_search(self):
        """
        Expect Table.filter to use match() not search(), so shouldn't match inside the
        middle of a cell.
        """

        self.assertEqual(
            0, len(self.table.filter(re.compile(r'developed regions$'))))

    def test_select(self):
        a = self.table.filter("WORLD")
        b = a.select(lambda t, b: t.y == b.y + 1 and t.x == b.x).value
        self.assertIn("More developed regions", b)

    def test_fill(self):
        a = self.table.filter("Variant")
        b = a.fill(xypath.LEFT).value
        self.assertEqual("Index", b)

    def test_shift(self):
        a = self.table.filter('Ethiopia')
        b = a.shift(-2, 2)  # down, left
        c = a.shift((-2, 2))  # as a tuple

        self.assertEqual(1, len(a))
        self.assertEqual(1, len(b))
        self.assertEqual(16.0, b.value)
        self.assertEqual(16.0, c.value)



import messytables
import xypath

def table_set(filename, *args, **kwargs):
    """get all the tables for a single spreadsheet"""
    with open(filename, 'rb') as f:
        mt_tableset = messytables.any.any_tableset(f, *args, **kwargs)
    return mt_tableset

def get_sheets(mt_tableset, ids):
    """get a subset of the tables from a tableset.
    Select them via a list or single item from:

    a string: tables with that name
    a number: the table with that index
    "*": get all of them
    a function: tables for which f(table) is truthy.
        i.e. you can use this with XYPath expressions.

    [5, 'kitten',
    lambda table: table.get_at(0,0).value == 'dog']
    will get you the 6th tab, all tabs named 'kitten'
    and all tabs where the top-left cell contains 'dog'."""

    def xy():
        xy_table = xypath.Table.from_messy(mt_table)
        xy_table.index = table_index
        return xy_table

    if isinstance(ids, (int, basestring)) or callable(ids):
        # it's a single thing, listify it
        ids = (ids, )

    for table_index, mt_table in enumerate(mt_tableset.tables):
        for identifier in ids:
            if identifier == '*':
                yield xy()
            elif isinstance(identifier, int):
                if identifier == table_index:
                    yield xy()
            elif isinstance(identifier, basestring):
                if identifier == mt_table.name:
                    yield xy()
            elif callable(identifier):
                xy_table = xy()
                if identifier(xy_table):
                    yield xy_table
            else:
                raise NotImplementedError("Don't know what to do with a {!r}".format(type(identifier)))



#!/usr/bin/env python
import sys
sys.path.append('xypath')
import xypath
import tcore


class TestJunctionMissing(tcore.TMissing):
    def test_cell_missing(self):
        a = self.table.filter("2").assert_one()
        b = self.table.filter("4").assert_one()
        junction_result = list(a.junction(b))
        self.assertEqual(len(junction_result), 0)


class TestJunction(tcore.TCore):
    def test_cell_junction(self):
        a = self.table.filter("WORLD").assert_one()
        b = self.table.filter("1990-1995").assert_one()
        junction_result = list(a.junction(b))
        self.assertEqual(1, len(junction_result))
        (x, y, z) = junction_result[0]
        self.assertIsInstance(x, xypath.Bag)
        self.assertIsInstance(y, xypath.Bag)
        self.assertIsInstance(z, xypath.Bag)
        self.assertEqual("WORLD", x.value)
        self.assertEqual("1990-1995", y.value)
        self.assertEqual(1.523, z.value)

    def test_bag_junction(self):
        a = self.table.filter("WORLD")
        b = self.table.filter("1990-1995")
        j = list(a.junction(b))
        self.assertEqual(1, len(j))
        (a_result, b_result, value_result) = j[0]
        self.assertEqual(1.523, value_result.value)

    def test_bag_junction_checks_type(self):
        bag = self.table.filter('Estimates')
        self.assertRaises(TypeError, lambda: list(bag.junction('wrong_type')))

    def test_junction_raises(self):
        a = self.table.filter('WORLD')
        b = self.table.filter('AFRICA')  # is below WORLD
        self.assertRaises(xypath.JunctionError, lambda: list(a.junction(b)))



import xypath
import xypath.xyzzy
from collections import OrderedDict

table = [[' ', 'Q', 'A', 'A', 'B', 'B'],
         ['K', 'W', 'a', 'b', 'a', 'b'],
         ['X', 'x', '5', '6', '7', '8'],
         ['X', 'y', '2', '9', '0', '1']]

output = [['A', 'a', 'x', '5'],
          ['A', 'a', 'y', '2'],
          ['A', 'b', 'x', '6'],
          ['A', 'b', 'y', '9'],
          ['B', 'a', 'x', '7'],
          ['B', 'a', 'y', '0'],
          ['B', 'b', 'x', '8'],
          ['B', 'b', 'y', '1']]


xy = xypath.Table.from_iterable(table)
dimensions_horizontal = [0, 1]
dimensions_vertical = [1]


def get_value_bags_for_dimension(table, header_index_and_direction, **kwargs):
    index, is_row = header_index_and_direction
    if is_row:
        function, direction = (lambda cell: cell.y == index, xypath.DOWN)
    else:
        function, direction = (lambda cell: cell.x == index, xypath.RIGHT)

    header = xypath.xyzzy.group(table.filter(function), **kwargs)
    return {k: header[k].fill(direction) for k in header}


def xtract_by_numbers(table, rows, columns):
    rows_and_columns = [(row_index, True) for row_index in rows] +\
                       [(column_index, False) for column_index in columns]
    bags_by_dimension = [get_value_bags_for_dimension(table, x) for x in rows_and_columns]

    def name_dim(index, dimension):
        return ('dimension_%s' % (index + 1), dimension)
    bags_for_xyzzy = OrderedDict(name_dim(i, x) for i, x in enumerate(bags_by_dimension))
    return xypath.xyzzy.xyzzy(table, bags_for_xyzzy, 'value')


def test_xyzzy_kinda_works():
    things = xtract_by_numbers(xy, dimensions_horizontal, dimensions_vertical)
    sorted_things = [thing.values() for thing in sorted(things)]
    assert sorted_things == output



#!/usr/bin/env python
import sys
sys.path.append('xypath')
import xypath
import tcore
from nose.tools import assert_raises

# Inheriting from tcore.TCore means that self.table contains an
# example table automatically, without us having to open a file
class TestTable(tcore.TCore):

    def test_get_at_gets_the_correct_cell(self):
        cells = self.table.get_at(4, 17)

        assert isinstance(cells, xypath.xypath.Bag)
        cells.assert_one() # will raise exception if more than one cell in bag
        assert cells.value == 900

    def test_get_at_returns_empty_bag_for_invalid_coordinates(self):
        cells = self.table.get_at(9999,9999)
        assert isinstance(cells, xypath.xypath.Bag)
        assert len(cells) == 0

    def test_get_at_complains_nicely(self):
        assert_raises(AssertionError, self.table.get_at, 'kitten')

    def test_table_has_name(self):
        assert self.table.name == u'ESTIMATES'
        copy = xypath.Table.from_bag(self.table)
        assert copy.name == u'ESTIMATES'



#!/usr/bin/env python
import sys
sys.path.append('xypath')
import xypath
import tcore
import unittest

class Test_Bag(tcore.TCore):
    def test_bag_from_list(self):
        "That Bag.from_list works and table is preserved"
        true_bag = self.table.filter(lambda b: b.value > "F")
        fake_bag = list(true_bag.table)
        self.assertEqual(type(fake_bag), list)
        remade_bag = xypath.Bag.from_list(fake_bag)
        self.assertEqual(true_bag.table, remade_bag.table)
        self.assertEqual(type(remade_bag), xypath.Bag)
        self.assertEqual(len(remade_bag), len(self.table))

    def test_bag_equality(self):
        lhs = self.table.filter("WORLD")
        rhs = self.table.filter("WORLD")

        self.assertEqual(lhs, rhs)

        rhs = self.table.filter("Variant")

        self.assertNotEqual(lhs, rhs)

    def test_bag_locations(self):
        world=self.table.filter("WORLD")
        empty=self.table.filter(lambda cell: False)

        self.assertEqual(world.excel_locations(), "C17")
        self.assertEqual(empty.excel_locations(), "")
        self.assertIn(", ...", self.table.excel_locations())

    def test_bags_from_different_tables_are_not_equal(self):
        bag1 = self.table.filter(lambda b: True)
        bag2 = xypath.Table.from_bag(bag1)
        self.assertNotEqual(bag1, bag2)

    def test_corebag_iterator_size(self):
        """Test that the iterator yields as many results as len() claims"""
        bag = self.table.filter('Estimates')
        self.assertEqual(265, len(bag))
        self.assertEqual(len(bag), len(list(bag)))

    def test_corebag_iterator_nonduplicate(self):
        """
        Ensure that each call of CoreBag.__iter__ returns a new iterator

        Supercedes _test_corebag_iterator_size_squared
        """

        bag = self.table.filter('Estimates')
        assert iter(bag) is not iter(bag)

    def _test_corebag_iterator_size_squared(self):
        """Worry: that iterating twice over bag doesn't work property.
        Test: that every pair of cells from the bags is present."""
        bag = self.table.filter('Estimates')

        SIZE = 4
        # Limit bag size so that test isn't slow
        bag = bag.from_list([cell for cell in bag][:SIZE])

        n_bag = len(bag)
        assert n_bag == SIZE

        count = 0
        for i in bag:
            for j in bag:
                count = count + 1

        self.assertEqual(count, n_bag * n_bag)

    def test_corebag_iterator_returns_bags(self):
        """Check the iterator returns bags, not _XYCells"""
        bag = self.table.filter('Estimates')
        for individual_cell in bag:
            self.assertIsInstance(individual_cell, xypath.Bag)

    def test_singleton_bag_value(self):
        self.assertEqual(
            "Country code",
            self.table.filter("Country code").value)
        self.assertRaises(xypath.XYPathError,
                          lambda: self.table.filter("Estimates").value)

    def test_messytables_has_properties(self):
        for bag in self.table.unordered:
            bag.properties.get("jam")  # is vaguely dict-like

    def test_from_bag(self):
        world_pops_bag = self.table.filter(lambda b: b.y >= 16 and
                                                     b.y <= 22 and
                                                     b.x >= 5 and
                                                     b.x <= 16)
        world_pops_table = xypath.Table.from_bag(world_pops_bag)

        # check extending the whole table gets lots of stuff all the way down
        fifties = self.table.filter("1950-1955")
        filties_col = fifties.fill(xypath.DOWN)
        self.assertEqual(265, len(filties_col))

        # check if we extend within the new world-only table, it only gets
        # stuff from the table
        fifties = world_pops_table.filter("1950-1955")
        filties_col = fifties.fill(xypath.DOWN)
        self.assertEqual(6, len(filties_col))

    def test_fill_down_without_termination(self):
        world_cell = self.table.filter('WORLD').assert_one()
        filled_down = world_cell.fill(xypath.DOWN)
        assert world_cell not in filled_down
        self.assertEqual(264, len(filled_down))

    def test_fill_right_without_termination(self):
        world_cell = self.table.filter('WORLD').assert_one()
        filled_right = world_cell.fill(xypath.RIGHT)
        assert world_cell not in filled_right
        self.assertEqual(14, len(filled_right))

    def test_bag_intersection(self):
        bag = self.table.filter('Estimates')
        another_bag = self.table.filter("WORLD")
        union = bag | another_bag
        intersection = union & another_bag
        assert intersection.value == "WORLD"

    def test_bag_union(self):
        bag = self.table.filter('Estimates')
        another_bag = self.table.filter("WORLD")

        union = bag | another_bag
        self.assertEqual(len(bag) + len(another_bag), len(union))

    def test_bag_set_difference(self):
        super_bag = self.table.filter("WORLD").assert_one().fill(xypath.DOWN)
        sub_bag = self.table.filter("AFRICA").assert_one()

        diff = super_bag - sub_bag

        self.assertEqual(len(super_bag) - len(sub_bag), len(diff))

    def test_bag_set_difference_rhs_non_subset_of_lhs(self):
        super_bag = self.table.filter("WORLD").assert_one().fill(xypath.DOWN)
        unrelated_bag = self.table.filter("Estimates")

        diff = super_bag - unrelated_bag

        self.assertEqual(len(super_bag), len(diff))
        self.assertEqual(super_bag, diff)

    def test_bag_ordering(self):


        # The purpose of this test is to have a "reasonably pathological" bag
        # so that the sort test is meaningful

        col1 = self.table.filter("Index").fill(xypath.DOWN)
        col2 = self.table.filter("Variant").fill(xypath.DOWN)

        bag = col1 | col2

        self.assertEqual([1.0, "Estimates", 2.0, "Estimates"],
                         [cell.value for cell in list(bag)[:4]])

        def yx(cell):
            return (cell.y, cell.x)

        self.assertEqual(sorted(bag, key=yx), list(bag))

    def test_assert_one_with_zero(self):
        bag = self.table.filter('No Such Cell')
        self.assertRaises(AssertionError, lambda: bag.assert_one())
        self.assertRaises(
            xypath.NoCellsAssertionError, lambda: bag.assert_one())

    def test_assert_one_with_multiple(self):
        bag = self.table.filter('Estimates')
        self.assertRaises(AssertionError, lambda: bag.assert_one())
        self.assertRaises(
            xypath.MultipleCellsAssertionError, lambda: bag.assert_one())

    def test_table_rows(self):
        counter = 0
        rows = list(self.table.rows())
        for row in rows:
            counter = counter + len(row)
        self.assertEqual(len(self.table), counter)  # misses none
        self.assertEqual(len(rows), 282)

    def test_table_cols(self):
        counter = 0
        cols = list(self.table.cols())
        for col in cols:
            counter = counter + len(col)
        self.assertEqual(len(self.table), counter)  # misses none
        self.assertEqual(len(cols), 17)

    def test_has_table(self):
        self.assertEqual(xypath.Table, type(self.table))
        self.assertIsInstance(self.table, xypath.Bag)

    def test_select_other(self):
        raise self.skipTest("select_other not tested")



#!/usr/bin/env python
import sys

import unittest
sys.path.append('xypath')
import xypath
import re
import tcore


class TestFilterOne(tcore.TCore):

    def test_filter_one_returns_a_single_cell(self):
        bag = self.table.filter_one('Variant')
        self.assertEquals(len(bag), 1)

    def test_filter_one_raises_error_if_no_cells_match(self):
        self.assertRaisesWithMessage(
            self.table.filter_one,
            xypath.NoCellsAssertionError,
            "We expected to find one cell containing the string 'I DO NOT EXIST', but we found none.",
            'I DO NOT EXIST'
        )

    def test_filter_one_raises_error_if_more_than_one_cell_matches(self):
        try:
            self.table.filter_one(re.compile(r'.*Europe.*'))
        except xypath.MultipleCellsAssertionError as e:
            pass
        else:
            assert False, "Didn't get a MultipleCellsAssertionError"

        assert "We expected to find one cell matching the regex '.*Europe.*', but we found 4: " in str(e)
        for cell in "C157, C146, C171, C188".split(', '):
            assert cell in str(e)



#!/usr/bin/env python

from collections import OrderedDict
from itertools import groupby
from xypath import Bag


class LookupFailureError(Exception):
    pass


def group(self, keyfunc=None):  # XYZZY
    """get a dictionary containing lists of singleton bags with the same
       value (by default; other functions available)"""
    groups = {}
    if keyfunc is None:
        keyfunc = lambda x: x.value
    protogroups = groupby(sorted(self, key=keyfunc), key=keyfunc)
    for k, v in protogroups:
        newbag = Bag.from_list(v)
        newbag.table = self.table
        groups[k] = newbag
    return groups


def headerheader(self, dir1, dir2, **kwargs):  # XYZZY
    """Given a header (e.g. "COUNTRY") get all things in one direction
       from it (e.g. down: "FRANCE", "GERMANY"), then use those to get
       a suitable xyzzy dict"""
    header = group(self.fill(dir1), **kwargs)
    return {k: header[k].fill(dir2) for k in header}


def are_distinct(fields):
    """True if the fields passed to xyzzy don't contain repeating cells."""
    allbags = set()
    bagcount = 0
    for field in fields:
        for bag in field.values():
            allbags = allbags.union(bag)
            bagcount = bagcount + len(bag)
    return len(allbags) == bagcount


def xyzzy(self, fields, valuename='_value'):  # XYZZY
    assert are_distinct(fields.values())
    fieldkeys = fields.keys()
    assert valuename not in fieldkeys

    def fieldlookup(cell):
        for fieldvalue in fields[field]:  # AFG, GBR; 1998, 1999
            if cell in fields[field][fieldvalue]:  # i.e. in bag
                return fieldvalue
        raise LookupFailureError

    for cell in self:
        path = OrderedDict()
        try:
            for field in fields:  # country, year
                path[field] = fieldlookup(cell)
        except LookupFailureError:
            continue

        path[valuename] = cell.value
        yield path



from setuptools import setup, find_packages

long_desc = """
XYPath is aiming to be XPath for spreadsheets: it offers a framework for
navigating around and extracting values from tabular data.
"""
# See https://pypi.python.org/pypi?%3Aaction=list_classifiers for classifiers

setup(
    name='xypath',
    version='1.0.7',
    description="Extract fields from tabular data with complex expressions.",
    long_description=long_desc,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
    ],
    keywords='',
    author='ScraperWiki Ltd',
    author_email='feedback@scraperwiki.com',
    url='http://scraperwiki.com',
    license='BSD',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=[],
    include_package_data=False,
    zip_safe=False,
    install_requires=[
        'messytables',
    ],
    tests_require=[],
    entry_points=\
    """
    """,
)



import unittest
from nose.tools import assert_raises, assert_equal
import xypath.contrib.excel as exceltool

class Test_Loader(unittest.TestCase):
    def test_excel_address(self):
        assert_equal(exceltool.excel_address_coordinate("A1"), (0, 0))
        assert_equal(exceltool.excel_address_coordinate("IV1"), (255, 0))
        assert_equal(exceltool.excel_address_coordinate("BGQCV1001"), (2**20-1, 1000))

    def test_invalid_excel_address(self):
        assert_raises(exceltool.InvalidExcelReference, exceltool.excel_address_coordinate, "CAT")

    def test_excel_col(self):
        assert_equal(exceltool.excel_column_number("A"), 1)
        assert_equal(exceltool.excel_column_number("Z"), 26)
        assert_equal(exceltool.excel_column_number("Z")+1, exceltool.excel_column_number("AA"))
        assert_equal(exceltool.excel_column_number("AZ")+1, exceltool.excel_column_number("BA"))
        assert_equal(exceltool.excel_column_number("ZZ")+1, exceltool.excel_column_number("AAA"))
        assert_equal(exceltool.excel_column_number("CAZ")+1, exceltool.excel_column_number("CBA"))

    def test_col_label(self):
        assert_equal(exceltool.excel_column_label(1), "A")
        assert_equal(exceltool.excel_column_label(256), "IV")
        assert_equal(exceltool.excel_column_label(2**20), "BGQCV")



import unittest
import xypath.loader
from nose.tools import assert_equal

class Test_Loader(unittest.TestCase):
    def test_foo(self):
        fn = "fixtures/4132.zip"
        tables = xypath.loader.table_set(fn)

        four, = xypath.loader.get_sheets(tables, 0)
        assert four.value == 'four', four

        one, three = xypath.loader.get_sheets(tables, [1, 2])
        assert one.value == 'one' and three.value == 'three', (one, three)

        all_sheets = xypath.loader.get_sheets(tables, 'table')
        assert len(list(all_sheets)) == 4

        all_sheets = xypath.loader.get_sheets(tables, '*')
        assert len(list(all_sheets)) == 4

        two, = xypath.loader.get_sheets(tables, lambda sheet: sheet.filter("two"))
        assert two.value == 'two'



#!/usr/bin/env python
import sys
import unittest
sys.path.append('xypath')
import xypath
import messytables
try:
    import hamcrest
except ImportError:
    hamcrest = None
import re
import tcore

class Test_Import_Missing(tcore.TMissing):
    def test_table_has_properties_at_all(self):
        self.table.sheet

class Test_Import(tcore.TCore):
    def test_table_has_sheet_properties(self):
        self.assertIn('xlrd', repr(self.table.sheet))

    #import
    def test_from_filename_with_table_name(self):
        """Can we specify only the filename and 'name' of the table?"""
        if hamcrest is None:
            raise unittest.SkipTest("Requires Hamcrest")
        table = xypath.Table.from_filename(
            self.wpp_filename,
            table_name='NOTES')
        self.assertEqual(32, len(table))
        table.filter(
            hamcrest.contains_string('(2) Including Zanzibar.')).assert_one()

    #import
    def test_from_filename_with_table_index(self):
        """Can we specify only the filename and index of the table?"""
        new_table = xypath.Table.from_filename(self.wpp_filename,
                                               table_index=5)
        self.assertEqual(1, len(new_table.filter('(2) Including Zanzibar.')))

    #import
    def test_from_file_object_table_index(self):
        with open(self.wpp_filename, 'rb') as f:
            extension = tcore.get_extension(self.wpp_filename)
            new_table = xypath.Table.from_file_object(
                f, extension, table_index=5)
        self.assertEqual(1, len(new_table.filter('(2) Including Zanzibar.')))

    #import
    def test_from_file_object_table_name(self):
        with open(self.wpp_filename, 'rb') as f:
            extension = tcore.get_extension(self.wpp_filename)
            new_table = xypath.Table.from_file_object(
                f, extension, table_name='NOTES')
        self.assertEqual(1, len(new_table.filter('(2) Including Zanzibar.')))

    #import
    def test_from_file_object_no_table_specifier(self):
        with open(self.wpp_filename, 'rb') as f:
            extension = tcore.get_extension(self.wpp_filename)
            self.assertRaises(
                TypeError,
                lambda: xypath.Table.from_file_object(f, extension))

    #import
    def test_from_file_object_ambiguous_table_specifier(self):
        with open(self.wpp_filename, 'rb') as f:
            extension = tcore.get_extension(self.wpp_filename)

            self.assertRaises(
                TypeError,
                lambda: xypath.Table.from_file_object(
                    f, extension, table_name='NOTES', table_index=4))

    #import
    def test_from_messy(self):
        new_table = xypath.Table.from_messy(self.messy.tables[0])
        self.assertEqual(265, len(new_table.filter('Estimates')))



#!/usr/bin/env python

""" musings on order of variables, x/y vs. col/row
Everyone agrees that col 2, row 1 is (2,1) which is xy ordered.
This works well with the name.
Remember that the usual iterators (over a list-of-lists)
is outer loop y first."""

import re
import messytables
import os
try:
    import hamcrest
    have_ham = True
except ImportError:
    have_ham = False

from collections import defaultdict
from copy import copy
from itertools import product, takewhile

import contrib.excel

UP = (0, -1)
RIGHT = (1, 0)
DOWN = (0, 1)
LEFT = (-1, 0)
UP_RIGHT = (1, -1)
DOWN_RIGHT = (1, 1)
UP_LEFT = (-1, -1)
DOWN_LEFT = (-1, 1)


class XYPathError(Exception):
    """Problems with spreadsheet layouts should raise this or a descendant."""
    pass


class JunctionError(RuntimeError, XYPathError):
    """Raised if paranoid _XYCell.junction finds it is returning one of the
       input cells - i.e. the input cells are in the same row or column"""
    pass


class NoCellsAssertionError(AssertionError, XYPathError):
    """Raised by Bag.assert_one() if the bag contains zero cells."""
    pass


class MultipleCellsAssertionError(AssertionError, XYPathError):
    """Raised by Bag.assert_one() if the bag contains multiple cells."""
    pass

class LookupConfusionError(AssertionError, XYPathError):
    """Lookup found multiple equally-close headers"""
    pass

class NoLookupError(AssertionError, XYPathError):
    """Lookup found no valid header"""
    pass

def describe_filter_method(filter_by):
        if callable(filter_by):
            return "matching a function called {}".format(filter_by.__name__)
        if isinstance(filter_by, basestring):
            return "containing the string {!r}".format(filter_by)
        if have_ham and isinstance(filter_by, hamcrest.matcher.Matcher):
            return "containing "+str(filter_by)
        if isinstance(filter_by, re._pattern_type):
            return "matching the regex {!r}".format(filter_by.pattern)
        else:
            return "which we're surprised we found at all"


class _XYCell(object):
    """needs to contain: value, position (x,y), parent bag"""
    __slots__ = ['value', 'x', 'y', 'table', 'properties']

    def __init__(self, value, x, y, table, properties=None):
        self.value = value  # of appropriate type
        self.x = x  # column number
        self.y = y  # row number
        self.table = table
        if properties is None:
            self.properties = {}
        else:
            self.properties = properties

    def __hash__(self):
        """
        In order to make a set of cells (used in Bag), they *must* be hashable.
        An _XYCell is uniquely identified (by sets, etc) through its position,
        content, and parent table.
        Note that `properties` is ignored since dicts are unhashable, and
        value may be redundant.
        """
        return hash((self.value, self.x, self.y, self.table))

    def __eq__(self, rhs):
        """See _XYCell.__hash__ for equality conditions"""
        return hash(self) == hash(rhs)

    def copy(self, new_table=None):
        """Make a copy of the cell.
           Its table will be new_table, if specified"""
        if new_table is None:
            new_table = self.table
        return _XYCell(self.value, self.x, self.y,
                       new_table, self.properties)

    def __repr__(self):
        return "_XYCell(%r, %r, %r)" % \
            (self.value, self.x, self.y)

    def __unicode__(self):
        return unicode(self.value)

    def lookup(self, header_bag, direction, strict=False):
        """
        Given a single cell (usually a value), a bag containing the headers
        of a particular type for that cell, and the direction in which to
        search for the relevant header

        e.g. for value cell V, searching up:

         [ ]                         [ ]
                                     [O]
                                                [ ]
                   ---> [ ]
                                      V
                                     [ ]
                             [ ]

        the cell with the arrow will be returned.

        Strict restricts the selection to cells in the same row/column as
        the value, so O is selected instead."""
        def mult(cell):
            return cell.x * direction[0] + cell.y * direction[1]

        def same_row_col(a, b, direction):
            return  (a.x - b.x  == 0 and direction[0] == 0) or \
                    (a.y - b.y  == 0 and direction[1] == 0)

        best_cell = None
        second_best_cell = None
        for target_cell in header_bag.unordered_cells:
            if mult(self) <= mult(target_cell):
                if not best_cell or mult(target_cell) <= mult(best_cell):
                    if not strict or same_row_col(self, target_cell, direction):
                        second_best_cell = best_cell
                        best_cell = target_cell
        if second_best_cell and mult(best_cell) == mult(second_best_cell):
            raise LookupConfusionError("{!r} is as good as {!r} for {!r}".format(
                best_cell, second_best_cell, self))
        if best_cell is None:
            raise NoLookupError("No lookup for {!r}".format(self))
        return best_cell


    def junction(self, other, direction=DOWN, paranoid=True):
        """ gets the lower-right intersection of the row of one, and the
        column of the other.

        paranoid: should we panic if we're hitting one of our input cells?"""

        def junction_coord(cells, direction=DOWN):
            """
            Under the hood: given two cells and a favoured direction, get the
            position of the cell with the column of one and the row of the
            other:

            A---->+
            |     ^
            |     |
            |     |
            v     |
            *<----B

            Both + and * are candidates for the junction of A and B - we take
            the one furthest down by default (specified by direction)

            >>> cells_dr = (_XYCell(0,1,2,None), _XYCell(0,3,4,None))
            >>> junction_coord(cells_dr, DOWN)
            (1, 4)
            >>> junction_coord(cells_dr, UP)
            (3, 2)
            >>> junction_coord(cells_dr, LEFT)
            (1, 4)
            >>> junction_coord(cells_dr, RIGHT)
            (3, 2)
            >>> cells_tr = (_XYCell(0,1,4,None), _XYCell(0,3,2,None))
            >>> junction_coord(cells_tr, DOWN)
            (3, 4)
            >>> junction_coord(cells_tr, UP)
            (1, 2)
            >>> junction_coord(cells_tr, LEFT)
            (1, 2)
            >>> junction_coord(cells_tr, RIGHT)
            (3, 4)
            """


            new_cells = (
                        (cells[0].x, cells[1].y),
                        (cells[1].x, cells[0].y)
            )
            for index, value in enumerate(direction):
                if value == 0:
                    continue
                if cmp(new_cells[0][index], new_cells[1][index]) == value:
                    return new_cells[0]
                else:
                    return new_cells[1]

        (x, y) = junction_coord((self, other), direction)
        if paranoid and (x, y) == (self.x, self.y) or \
                        (x, y) == (other.x, other.y):
            raise JunctionError(
                "_XYCell.junction(_XYCell) resulted in a cell which is equal"
                " to one of the input cells.\n"
                "  self: {}\n  other: {}\n  x: {}\n  y: {}".format(
                    self, other, x, y))
        junction_bag = self.table.get_at(x, y)
        if len(junction_bag) == 0:
            return
        self_bag = Bag(self.table)
        self_bag.add(self)
        other_bag = Bag(self.table)
        other_bag.add(other)
        yield (self_bag, other_bag, junction_bag)

    def shift(self, x=0, y=0):
        """Get the cell which is offset from this cell by x columns, y rows"""
        if not isinstance(x, int):
            assert y == 0, \
                "_XYCell.shift: x=%r not integer and y=%r specified" % (x, y)
            return self.shift(x[0], x[1])
        return self.table.get_at(self.x + x, self.y + y)._cell


class CoreBag(object):
    """Has a collection of _XYCells"""
    def pprint(self, *args, **kwargs):
        return contrib.excel.pprint(self, *args, **kwargs)

    def as_list(self, *args, **kwargs):
        return contrib.excel.as_list(self, *args, **kwargs)

    def filter_one(self, filter_by):
        return contrib.excel.filter_one(self, filter_by)

    def excel_locations(self, *args, **kwargs):
        return contrib.excel.excel_locations(self, *args, **kwargs)

    def __init__(self, table):
        self.__store = set()
        self.table = table

    def add(self, cell):
        """Add a cell to this bag"""
        if not isinstance(cell, _XYCell):
            raise TypeError("Can only add _XYCell types to Bags: {}".format(
                            cell.__class__))
        self.__store.add(cell)

    def __eq__(self, other):
        """Compare two bags: they are equal if:
           * their table are the same table (object)
           * they contain the same set of cells"""
        if not isinstance(other, CoreBag):
            return False
        return (self.table is other.table and
                self.__store == other.__store)

    def __len__(self):
        return len(self.__store)

    def __repr__(self):
        return repr(self.__store)

    @classmethod
    def singleton(cls, cell, table):
        """
        Construct a bag with one cell in it
        """
        bag = cls(table=table)
        bag.add(cell)
        return bag

    @property
    def unordered(self):
        """
        Obtain an unordered iterator over this bag. iter(bag) is sorted on
        demand, and therefore inefficient if being done repeatedly where order
        does not matter.
        """
        return (Bag.singleton(c, table=self.table) for c in self.__store)

    @property
    def unordered_cells(self):
        """
        Analogous to the `unordered` property, except that it returns _XYCells
        instead of Bags.
        """
        return iter(self.__store)

    def __iter__(self):
        """
        Return a view of the cells in this back in left-right, top-bottom order
        Note: this is expensive for large bags (when done repeatedly). If you
        don't care about order, use `bag.unordered`, which gives an unordered
        iterator.
        """
        def yx(cell):
            return cell.y, cell.x

        for cell in sorted(self.__store, key=yx):
            yield Bag.singleton(cell, table=self.table)

    def __sub__(self, rhs):
        """Bags quack like sets. Implements - operator."""
        return self.difference(rhs)

    def difference(self, rhs):
        """Bags quack like sets."""
        assert self.table is rhs.table,\
            "Can't difference bags from separate tables"
        new = copy(self)
        new.__store = self.__store.difference(rhs.__store)
        return new

    def __or__(self, rhs):
        """Bags quack like sets. Implements | operator.
           For mathematical purity, + (__add__) isn't appropriate"""
        return self.union(rhs)

    def union(self, rhs):
        """Bags quack like sets."""
        assert self.table is rhs.table, "Can't union bags from separate tables"
        new = copy(self)
        new.__store = self.__store.union(rhs.__store)
        return new

    def __and__(self, rhs):
        return self.intersection(rhs)

    def intersection(self, rhs):
        assert self.table is rhs.table, \
            "Can't take intersection of bags from separate tables"
        new = copy(self)
        new.__store = self.__store.intersection(rhs.__store)
        return new

    def select(self, function):
        """Select cells from this bag's table based on the cells in this bag.
            e.g.
            bag.select(lambda bag_cell, table_cell: bag_cell.y == table_cell.y
                and bag_cell.value == table_cell.value)
            would give cells in the table with the same name on the same row
            as a  cell in the bag"""
        return self.table.select_other(function, self)

    def select_other(self, function, other):
        """A more general version of select, where another bag to select from
        is explicitly specified rather than using the original bag's table"""
        """note: self.select(f) = self.table.select_other(f, self)"""
        newbag = Bag(table=self.table)
        for bag_cell in self.__store:
            for other_cell in other.__store:
                if function(bag_cell, other_cell):
                    newbag.add(bag_cell)
                    break
        return newbag

    def filter(self, filter_by):
        """
        Returns a new bag containing only cells which match
        the filter_by predicate.

        filter_by can be:
        a) a callable, which takes a cell as a parameter
           and returns True if the cell should be returned,
           such as `lambda cell: cell value == 'dog'
        b) a string, to match exactly: `u'dog'`
        c) a hamcrest match rule: `hamcrest.equal_to("dog")
           (requires hamcrest to be available)
        d) a compiled regex: `re.compile("dog")
        """
        if callable(filter_by):
            return self._filter_internal(filter_by)
        elif isinstance(filter_by, basestring):
            return self._filter_internal(lambda cell: unicode(cell.value).strip() == filter_by)
        elif have_ham and isinstance(filter_by, hamcrest.matcher.Matcher):
            return self._filter_internal(lambda cell: filter_by.matches(cell.value))
        elif isinstance(filter_by, re._pattern_type):
            return self._filter_internal(
                lambda cell: re.match(filter_by, unicode(cell.value)))
        else:
            raise ValueError("filter_by must be function, hamcrest filter, compiled regex or string.")

    def _filter_internal(self, function):
        newbag = Bag(table=self.table)
        for bag_cell in self.unordered_cells:
            if function(bag_cell):
                newbag.add(bag_cell)
        return newbag

    def assert_one(self, message="assert_one() : {} cells in bag, not 1"):
        """Chainable: raise an error if the bag contains 0 or 2+ cells.
           Otherwise returns the original (singleton) bag unchanged."""
        if len(self.__store) == 1:
            return self

        elif len(self.__store) == 0:
            raise NoCellsAssertionError(
                message.format(
                    len(self.__store)
                )
            )

        elif len(self.__store) > 1:
            raise MultipleCellsAssertionError(
                message.format(
                    len(self.__store)
                )
            )

    @property
    def _cell(self):
        """Under the hood: get the cell inside a singleton bag.
           It's an error for it to not contain precisely one cell."""
        try:
            xycell = list(self.assert_one().__store)[0]
        except AssertionError:
            l = len(list(self.__store))
            raise XYPathError("Can't use multicell bag as cell: (len %r)" % l)
        else:
            assert isinstance(xycell, _XYCell)
            return xycell

    @property
    def value(self):
        """Getter for singleton's cell value"""
        return self._cell.value

    @property
    def x(self):
        """Getter for singleton's cell column number"""
        return self._cell.x

    @property
    def y(self):
        """Getter for singleton's cell row number"""
        return self._cell.y

    @property
    def properties(self):
        """Getter for singleton's cell properties"""
        return self._cell.properties



class Bag(CoreBag):

    @staticmethod
    def from_list(cells):
        """
        Make a non-bag iterable of cells into a Bag. Some magic may be lost,
        especially if it's zero length.
        TODO: This should probably be part of the core __init__ class.
        TODO: Don't do a piece-by-piece insertion, just slap the whole listed
              iterable in, because this is slow.
        """  # TODO
        bag = Bag(table=None)
        for i, cell_bag in enumerate(cells):
            bag.add(cell_bag._cell)
            if i == 0:
                bag.table = cell_bag.table
            else:
                assert bag.table == cell_bag.table
        return bag

    def fill(self, direction, stop_before=None):
        """Should give the same output as fill, except it
        doesn't support non-cardinal directions or stop_before.
        Twenty times faster than fill in test_ravel."""
        if direction in (UP_RIGHT, DOWN_RIGHT, UP_LEFT,
                                        UP_RIGHT):
            return self._fill(direction, stop_before)

        def what_to_get(cell):
            """converts bag coordinates into thing to pass to get_at"""
            cell_coord = (cell.x, cell.y)
            retval = []
            for cell_coord, direction_coord in zip(cell_coord, direction):
                if direction_coord != 0:
                    retval.append(None)
                else:
                    retval.append(cell_coord)
            return tuple(retval)  # TODO yuck

        if direction not in (UP, RIGHT, DOWN, LEFT):
            raise ValueError("Must be a cardinal direction!")

        ### this is what same_row/col should look like!
        small_table = None
        for cell in self.unordered_cells:
            got_rowcol = self.table.get_at(*what_to_get(cell))
            if small_table:
                small_table = small_table.union(got_rowcol)
            else:
                small_table = got_rowcol

        # now we use the small_table as if it was the table.
        (left_right, up_down) = direction
        bag = small_table.select_other(
            lambda table, bag: cmp(table.x, bag.x) == left_right
            and cmp(table.y, bag.y) == up_down,
            self
        )
        if stop_before is not None:
            return bag.stop_before(stop_before)
        else:
            return bag

    def stop_before(self, stop_function):
        """Assumes the data is:
           * in a single row or column
           * proceeding either downwards or rightwards
        """
        return Bag.from_list(list(
            takewhile(lambda c: not stop_function(c), self)))

    def _fill(self, direction, stop_before=None):
        """
        If the bag contains only one cell, select all cells in the direction
        given, excluding the original cell. For example, from a column heading
        cell, you can "fill down" to get all the values underneath it.

        If you provide a stop_before function, it will be called on each cell
        as a stop condition. For example, if you provide a stop_before function
        which tests cell.value for an empty string. This would stop the fill
        function before it reaches the bottom of the sheet, for example.
        """
        raise DeprecationWarning("2D fill is deprecated. Yell if you need it.")
        if direction not in (UP, RIGHT, DOWN, LEFT, UP_RIGHT, DOWN_RIGHT,
                             UP_LEFT, DOWN_LEFT):
            raise ValueError("Invalid direction! Use one of UP, RIGHT, "
                             "DOWN_RIGHT etc")

        (left_right, up_down) = direction
        bag = self.select(
            lambda table, bag: cmp(table.x, bag.x) == left_right
            and cmp(table.y, bag.y) == up_down
        )

        if stop_before is not None:
            # NOTE(PMF): stop_before is limited to singleton bags, in the DOWN
            # or RIGHT direction. This isn't ideal, but with the above "magic"
            # cmp code I can't think of an elegant general way of doing this. I
            # also can't imagine what it means to run fill in multiple
            # directions, or with non singleton bags. TODO: Constrain?

            if direction not in (DOWN, RIGHT):
                raise ValueError("Oops, stop_before only works down or right!")
            self.assert_one("You can't use stop_before for bags with more than"
                            " one cell inside.")

            return Bag.from_list(list(
                takewhile(lambda c: not stop_before(c), bag)))

        return bag

    def junction(self, other, *args, **kwargs):
        """For every combination of pairs of cells from this bag and the other
           bag, get the cell that is at the same row as one of them, and column
           as the other.
           There are two: so we specify a direction to say which one wins (in
           the cell-based version of this function) - defaulting to the one
           furthest down"""
        if not isinstance(other, CoreBag):
            raise TypeError(
                "Bag.junction() called with invalid type {}, must be "
                "(Core)Bag".format(other.__class__))

        # Generate ordered lists of dimension cells exactly once (avoid doing
        # it in the inner loop because of the sorted() in __iter__)
        self_cells = list(self)
        other_cells = list(other)

        for self_cell in self_cells:
            for other_cell in other_cells:

                assert self_cell._cell.__class__ == other_cell._cell.__class__

                for triple in self_cell._cell.junction(other_cell._cell,
                                                       *args, **kwargs):
                    yield triple

    def shift(self, x=0, y=0):
        """
        Return a bag in which each cell is offset from the source bag by the
        coordinates specified. Coordinates can be specified as:
        Bag.shift(0,2) - full specification
        Bag.shift(y=2) - partial specification
        Bag.shift((0,2)) - use of tuple for x, unspecified y
        """
        if not isinstance(x, int):
            assert y == 0, \
                "Bag.shift: x=%r not integer and y=%r specified" % (x, y)
            return self.shift(x[0], x[1])
        bag = Bag(table=self.table)
        for b_cell in self.unordered:
            t_cell = self.table.get_at(b_cell.x + x, b_cell.y + y).assert_one()
            bag.add(t_cell._cell)
        return bag

    def extrude(self, dx, dy):
        """
        Extrude all cells in the bag by (dx, dy), by looking

        For example, given the bag with a cell at (0, 0):

            {(0, 0)}

        .extrude(2, 0) gives the bag with the cells (to the right):

            {(0, 0), (1, 0), (2, 0)}

        .extrude(0, -2) gives the bag with the cells (up):

            {(0, 0), (0, -1), (0, -2)}

        """

        if dx < 0:
            dxs = range(0, dx - 1, -1)
        else:
            dxs = range(0, dx + 1, +1)

        if dy < 0:
            dys = range(0, dy - 1, -1)
        else:
            dys = range(0, dy + 1, +1)

        bag = Bag(table=self.table)
        for cell in self.unordered_cells:
            for i, j in product(dxs, dys):
                bag.add(self.table.get_at(cell.x + i, cell.y + j)._cell)

        return bag

    def same_row(self, bag):
        """
        Select cells in this bag which are in the same
        row as a cell in the other `bag`.
        """
        # TODO: make less crap - use Table.get_at()
        all_y = set()
        for cell in bag.unordered_cells:
            all_y.add(cell.y)
        return self.filter(lambda c: c.y in all_y)

    def same_col(self, bag):
        """
        Select cells in this bag which are in the same
        row as a cell in the other `bag`.
        """
        # TODO: make less crap
        all_x = set()
        for cell in bag.unordered_cells:
            all_x.add(cell.x)
        return self.filter(lambda c: c.x in all_x)

    def __getattr__(self, name):
        if name.startswith("is_not_"):
            return self.filter(lambda cell: not cell.properties[name[7:]])
        if name.startswith("is_"):  # might need additional layer of indirection
            return self.filter(lambda cell: cell.properties[name[3:]])
        if name.endswith("_is_not"):
            return lambda value: self.filter(lambda cell: not cell.properties[name[:-7]] == value)
        if name.endswith("_is"):
            return lambda value: self.filter(lambda cell: cell.properties[name[:-3]] == value)
        raise AttributeError, "Bag has no attribute {!r}".format(name)


class Table(Bag):
    """A bag which represents an entire sheet.
       Features indices to speed retrieval by coordinate.
       Also includes functions for importing tables into XYPath"""
    def __init__(self, name=""):
        super(Table, self).__init__(table=self)
        self._x_index = defaultdict(lambda: Bag(self))
        self._y_index = defaultdict(lambda: Bag(self))
        self._max_x = -1
        self._max_y = -1
        self.sheet = None
        self.name = name

    def rows(self):
        """Get bags containing each row's cells, in order"""
        for row_num in range(0, self._max_y + 1):  # inclusive
            yield self._y_index[row_num]

    def cols(self):
        """Get bags containing each column's cells, in order"""
        for col_num in range(0, self._max_x + 1):  # inclusive
            yield self._x_index[col_num]

    def col(self, column):
        if isinstance(column, basestring):
            c_num = contrib.excel.excel_column_number(column, index=0)
            return self.col(c_num)
        else:
            assert isinstance(column, int)
            return self._x_index[column]

    def add(self, cell):
        """Under the hood: add a cell to a table and the table's indices.
           Used in the construction of a table."""
        self._x_index[cell.x].add(cell)
        self._y_index[cell.y].add(cell)
        self._max_x = max(self._max_x, cell.x)
        self._max_y = max(self._max_y, cell.y)
        super(Table, self).add(cell)

    def get_at(self, x=None, y=None):
        """Directly get a singleton bag via indices. Faster than Bag.filter"""
        # we use .get() here to avoid new empty Bags being inserted
        # into the index stores when a non-existant coordinate is requested.
        assert isinstance(x, int) or x is None, "get_at takes integers (got {!r})".format(x)
        assert isinstance(y, int) or y is None, "get_at takes integers (got {!r})".format(y)
        if x is None and y is None:
            raise TypeError('get_at requires at least one x or y value')
        if x is None:
            return self._y_index.get(y, Bag(self))
        if y is None:
            return self._x_index.get(x, Bag(self))
        return self._y_index.get((y), Bag(self)).filter(lambda cell: cell.x==x)

    @staticmethod
    def from_filename(filename, table_name=None, table_index=None):
        """Wrapper around from_file_object to handle extension extraction"""
        # NOTE: this is a messytables table name
        extension = os.path.splitext(filename)[1].strip('.')
        with open(filename, 'rb') as f:
            return Table.from_file_object(f, extension,
                                          table_name=table_name,
                                          table_index=table_index)

    @staticmethod
    def from_file_object(fobj, extension='',
                         table_name=None, table_index=None):
        """Load table from file object, you must specify a table's name
           or position number. If you don't know these, try from_messy."""
        # NOTE this is a messytables table name
        if (table_name is not None and table_index is not None) or \
                (table_name is None and table_index is None):
            raise TypeError("Must give exactly one of table_name, table_index")

        table_set = messytables.any.any_tableset(fobj, extension=extension)

        if table_name is not None:
            return Table.from_messy(table_set[table_name])
        elif table_index is not None:
            return Table.from_messy(table_set.tables[table_index])

    @staticmethod
    def from_messy(messy_rowset):
        """Import a rowset (table) from messytables, e.g. to work with each
           table in turn:
               tables = messytables.any.any_tableset(fobj)
               for mt_table in tables:
                   xy_table = xypath.Table.from_messy(mt_table)
                   ..."""

        assert isinstance(messy_rowset, messytables.core.RowSet),\
            "Expected a RowSet, got a %r" % type(messy_rowset)
        new_table = Table.from_iterable(
            messy_rowset,
            value_func=lambda cell: cell.value,
            properties_func=lambda cell: cell.properties,
            name=messy_rowset.name)

        if hasattr(messy_rowset, 'sheet'):
            new_table.sheet = messy_rowset.sheet
        return new_table

    @staticmethod
    def from_iterable(table, value_func=lambda cell: cell,
                      properties_func=lambda cell: {},
                      name=None):
        """Make a table from a pythonic table structure.
           The table must be an iterable which returns rows (in top-to-bottom
           order), which in turn are iterables which returns cells (in
           left-to-right order).
           value_func and properties_func specify how the cell maps onto an
           _XYCell's value and properties. The defaults assume that you have a
           straight-forward list of lists of values."""
        new_table = Table(name=name)
        for y, row in enumerate(table):
            for x, cell in enumerate(row):
                new_table.add(
                    _XYCell(
                        value_func(cell),
                        x,
                        y,
                        new_table,
                        properties_func(cell)))
        return new_table

    @staticmethod
    def from_bag(bag, name=None):
        """Make a copy of a bag which is its own table.
           Useful when a single imported table is two logical tables"""
        if name is None:
            name=bag.table.name
        new_table = Table(name=name)
        for bag_cell in bag.unordered:
            new_table.add(bag_cell._cell.copy(new_table))
        return new_table



import sys
from itertools import product
from ..extern.tabulate import tabulate
import xypath
import re

class InvalidExcelReference(Exception):
    pass

def excel_address_coordinate(address):
    """Given a cell reference, return a tuple suitable for inserting into Table.get_at()"""
    match = re.match("([A-Za-z]+)([0-9]+)$", address)
    if not match:
        raise InvalidExcelReference(address)
    row_name, col_num = match.groups()
    return (excel_column_number(row_name, index=0), int(col_num)-1)


def excel_column_number(raw_column_name, index=1):
     """Given a column name, give me the column number
     A=1
     Z=26
     AA=27
     AZ=52
     BA=53 etc"""

     def char_value(c):
         return ord(c)-64

     value = 0
     for char in raw_column_name.upper():
         value = value * 26
         value = value + char_value(char)
     return value - 1 + index

def excel_column_label(n):

    """
    Excel's column counting convention, counting from A at n=1
    """
    def inner(n):
        if n <= 0:
            return []
        if not n:
            return [0]
        div, mod = divmod(n - 1, 26)
        return inner(div) + [mod]
    return "".join(chr(ord("A") + i) for i in inner(n))


def excel_locations(bag, limit=3):
    """describe the locations of cells in a bag
       in Excel notation"""
    def _excel_location(cell):
        excelcol = excel_column_label(cell.x + 1)
        assert excelcol
        return "{}{}".format(excelcol,
                             cell.y)
    builder = []
    for i, singleton in enumerate(bag.unordered):
        if i >= limit:
            builder.append("...")
            break
        builder.append(_excel_location(singleton))
    return ', '.join(builder)


def filter_one(self, filter_by):
    errormsg = "We expected to find one cell {}, but we found {}."
    foundmsg = 'one'
    filtered = self.filter(filter_by)
    filter_length = len(filtered)

    if filter_length == 0:
        foundmsg = 'none'
    elif filter_length > 1:
        foundmsg = "{}: {}".format(
            filter_length,
            filtered.excel_locations(filtered)
        )

    return filtered.assert_one(
        errormsg.format(
            xypath.describe_filter_method(filter_by),
            foundmsg
        )
    )


def as_list(self, collapse_empty=False, excel_labels=False):
    """
    Generate a rectangular 2D list representing the bag, where the contents
    are the values of the cells, or None if there is no cell present at a
    particular location.

    collapse_empty: Remove empty rows and columns for a more compact
        representation

    excel_labels: Put "excel-like" row/column labels around the table
    """
    cells = list(self)
    cxs = [cell.x for cell in cells]
    cys = [cell.y for cell in cells]
    xmin, xmax = min(cxs), max(cxs)
    ymin, ymax = min(cys), max(cys)

    width, height = xmax-xmin+1, ymax-ymin+1

    result = [[None]*width for i in xrange(height)]

    for x, y in product(xrange(xmin, xmax+1), xrange(ymin, ymax+1)):
        cell = self.table.get_at(x, y)
        result[y-ymin][x-xmin] = cell.value if cell in self else None

    # Indices of the rows/columns that are present, starting at (1, 1)
    row_indices = list(range(ymin+1, ymax+1+1))
    col_indices = list(range(xmin+1, xmax+1+1))

    if collapse_empty:
        # Note the dreaded "delete from thing you're iterating over"
        # pattern here. Hence `reversed`.

        # Remove empty rows
        for j, y in reversed(list(enumerate(result))):
            if not any(y):
                del result[j]
                del row_indices[j]

        search_indices = list(range(width))
        # Find empty columns
        for y in result:
            for i, s_index in reversed(list(enumerate(search_indices))):
                if y[s_index]:
                    # We've found a thing. Don't need to search it anymore.
                    del search_indices[i]
        # A search_index making it here, is empty, delete it from all rows
        for y in result:
            for i in reversed(search_indices):
                del y[i]

        # Fix the set of column indices which remain
        for i in reversed(search_indices):
            del col_indices[i]

    if excel_labels:
        # Add row numbers to each row
        for j, row in zip(row_indices, result):
            row[0:0] = [j]

        # Add column labels to the headers
        header = [None] + [excel_column_label(i) for i in col_indices]
        result[0:0] = [header]

    return result


def pprint(self, collapse_empty=False,
           excel_labels=True, stream=sys.stdout):
    result = self.as_list(collapse_empty, excel_labels)

    for row in result:
        for i, cell in enumerate(row):
            if cell is None:
                cell = "/"
            cell = str(cell)
            row[i] = cell

    if excel_labels:
        print >>stream, tabulate(result[1:], headers=result[0])
    else:
        print >>stream, tabulate(result)
